/*
 * $Id$
 * Author: Karoly Bosa
 *
 * (C) 2008 by Research Institute for Symbolic Computation, Linz, Austria
 * All rights reserved. See COPYRIGHT in top-level directory.
 */

/* Description: This program establishes a tree structure of processes where NR_OF_TASKS tasks
 * are distributed (via the non-leaf nodes), elaborated (by the leaf nodes) and finally
 * the results of task are collected by the root (via the non-leaf nodes again).
 * The program works in case of tree structures with 2, 3, 4, ...any levels depending what
 * is in the xml-based description file.
 *
 * To compile this file use the Makefile located in the directory "examples"
 * and type the command "make tree"
 *
 * This program requires the following:
 *
 *  - a globus user credentials, you can type:
 * 		grid-proxy-init -verify -debug
 *
 * 	- a machine file with the name "machines" in the same directory
 *    with the following content at least:
 *		"/your_hostname/" 20
 *
 *	- finally the program assumes that each non-leaf node has AT LEAST TWO children
 *	  (described in the xml-based description).
 *
 * To execute this test using mpirun you can type e.g.:
 * 		mpirun -np 20 taagTree treeWith3levelsOnlocalhost.xml
 *
 * 			OR
 *
 * 		../../bin/taagrun treeWith3levelsOnlocalhost.xml
 *
 * Remark: In the second case machine file is not required.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mpi.h>
#include <taag.h>

#define MSG_SIZE 160
#define NR_OF_TASKS  101    //500
#define BUFFER_SIZE NR_OF_TASKS+2
#define EXIT_SIGNAL "EXIT"

#define XML_DEFAULT "treeWith3levelsOnlocalhost.xml"

void create_task(int i, char* s) {
	sprintf(s,"TASK%d",i);
}

void process_task(int rank, char* s, char* t) {
	int group_rank;
	char hname[MPI_MAX_PROCESSOR_NAME];
	int l;

	TAAG_Group_rank(rank, &group_rank);
	MPI_Get_processor_name(hname, &l);
	sprintf(t, "%s is processed by p.%d of g.%d on %s.", s, rank, group_rank, hname);

}

int main(int argc, char *argv[]) {
	int nrProcs, rc, flag, nrChildren;
	int rank, root, parent;

	int children[100];
	char outbuff[BUFFER_SIZE][MSG_SIZE];
	char inbuff[BUFFER_SIZE][MSG_SIZE];

	int indx_recv = 0;
	int child_indx = 0;
	int indx = 0;

	MPI_Request reqs[BUFFER_SIZE];
	MPI_Status stats[BUFFER_SIZE];
	MPI_Request req;
	MPI_Status stat;

	int triggeredExit = TAAG_FALSE;

	rc = MPI_Init(&argc,&argv);
	if (rc != MPI_SUCCESS) {
		printf("Error starting MPI program.\n");
		MPI_Abort(MPI_COMM_WORLD, rc);
	}
	MPI_Comm_size(MPI_COMM_WORLD,&nrProcs);
	MPI_Comm_rank(MPI_COMM_WORLD,&rank);

	if (argc > 1) {
		rc = TAAG_Init(argv[1]);
	}
	else {
		rc = TAAG_Init(XML_DEFAULT);
	}


	if (rc != TAAG_SUCCESS) {
		fprintf(stderr,"Error (code %d) initializing the TAAG structure on process %d.\n", rc, rank);
		MPI_Finalize();
		return rc;
	} //if

	TAAG_Tree_isTree(&flag);
	if (!flag) {
		fprintf(stderr, "Error (code %d) the given schema is NOT a tree.\n", TAAG_ERR_SCHEMA);
		TAAG_Free();
		MPI_Finalize();
		return TAAG_ERR_SCHEMA;
	}

	TAAG_Tree_root(&root);

	if (rank == root) {
		/***** root branch *****/
		TAAG_Tree_width(rank, 1, &nrChildren);
		TAAG_Tree_children(rank, 1, nrChildren, children);

		for (int i = 0; i < NR_OF_TASKS; i++) {
			create_task(i, outbuff[i]);
			MPI_Irecv(inbuff[i], MSG_SIZE, MPI_CHAR, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &reqs[i]);
			MPI_Isend(outbuff[i], MSG_SIZE, MPI_CHAR, children[child_indx], i, MPI_COMM_WORLD, &req);
			child_indx++;
			if (child_indx == nrChildren) child_indx = 0;
		}

		MPI_Waitall (NR_OF_TASKS, reqs, stats);
		for (int i = 0; i < NR_OF_TASKS; i++) {
		  printf("%s\n", inbuff[i]);
		} //for

		strcpy(outbuff[NR_OF_TASKS], EXIT_SIGNAL);
		for (int i = 0; i < nrChildren; i++) {
			MPI_Send(outbuff[NR_OF_TASKS], MSG_SIZE, MPI_CHAR, children[i], 0, MPI_COMM_WORLD);
		}
	} //if
	  else {
		/***** non-root branch *****/
		TAAG_Tree_parent(rank, &parent);
		TAAG_Tree_isLeaf(rank, &flag);
		if (flag == TAAG_FALSE) {
			/***** non-leaf branch *****/
			TAAG_Tree_width(rank, 1, &nrChildren);
			TAAG_Tree_children(rank, 1, nrChildren, children);

			MPI_Irecv(inbuff[indx_recv], MSG_SIZE, MPI_CHAR, parent, MPI_ANY_TAG, MPI_COMM_WORLD, &reqs[indx_recv]);
			while(!triggeredExit) {
				MPI_Wait(&reqs[indx_recv], &stat);
				indx = indx_recv;
				indx_recv++;
				if (!strcmp(inbuff[indx], EXIT_SIGNAL)) {
					triggeredExit = TAAG_TRUE;
					for (int i=0; i < nrChildren; i++) {
						MPI_Send(inbuff[indx], MSG_SIZE, MPI_CHAR, children[i], stat.MPI_TAG, MPI_COMM_WORLD);
					} //for
				} //if
				else {
					MPI_Irecv(inbuff[indx_recv], MSG_SIZE, MPI_CHAR, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &reqs[indx_recv]);
					if (stat.MPI_SOURCE == parent) {
						MPI_Isend(inbuff[indx], MSG_SIZE, MPI_CHAR, children[child_indx], stat.MPI_TAG, MPI_COMM_WORLD, &req);
						child_indx++;
						if (child_indx == nrChildren) child_indx = 0;
					}
					else {
						MPI_Isend(inbuff[indx], MSG_SIZE, MPI_CHAR, parent, stat.MPI_TAG, MPI_COMM_WORLD, &req);
					} //else
				} //else
			} //while
		} //if
		else {
			/***** leaf branch *****/
			int task_counter = 0;   //JUST FOR TESTING
			MPI_Irecv(inbuff[indx_recv], MSG_SIZE, MPI_CHAR, parent, MPI_ANY_TAG, MPI_COMM_WORLD, &reqs[indx_recv]);
			while(!triggeredExit) {
				MPI_Wait(&reqs[indx_recv], &stat);
				if (!strcmp(inbuff[indx_recv], EXIT_SIGNAL)) {
					printf("The number of processed tasks on process %d is %d.\n", rank, task_counter); //JUST FOR TESTING
					triggeredExit = TAAG_TRUE;
				} //if
				else {
					task_counter++; //JUST FOR TESTING
					indx = indx_recv;
					indx_recv++;
					MPI_Irecv(inbuff[indx_recv], MSG_SIZE, MPI_CHAR, parent, MPI_ANY_TAG, MPI_COMM_WORLD, &reqs[indx_recv]);
					process_task(rank, inbuff[indx], outbuff[indx]);
					MPI_Isend(outbuff[indx], MSG_SIZE, MPI_CHAR, parent, stat.MPI_TAG, MPI_COMM_WORLD, &req);
				} //else
			} //while
		} //else
	} //else
	TAAG_Free();
	MPI_Finalize();
	return 0;
}

